#!/usr/bin/perl

# YWMS server
#
# emulates a primitive WMS server (only GetMap requests) that serves Yahoo! satellite
# images.
#
# For each incoming request,
# * creates a temporary "web page" requesting the required image from Yahoo!
# * starts an Xvfb server to display the browser on
# * starts a firefox instance to render the web page onto the Xvfb display
# * converts the resulting image to jpeg and returns it
#
# This method can theoretically be used to display maps from any web site where the
# API supports selecting a geographic region AND includes a call to find out the 
# region actually displayed.
#
# This is required because most such services have discrete zoom levels, while the
# WMS request issued by josm expects to receive EXACTLY the coordinates requested
# in an image of the requested pixel size.
#
# The temporary HTML/Javscript generated by this server uses the Yahoo! API to display
# an image depicting the selected area in the best available zoom level and then 
# inquires about the area actually displayed, which will always be larger than what
# was requested. This information is then written to stdout using the "dump" Javascript
# command. From there it is read by this server, and used to stretch and cut the 
# resulting browser image to the size requested in the WMS request. 
#
# To illustrate:
#
# * josm says "I want the area (48.5,7.0)-(48.8,7.1) as a 1000x800 pixel image"
# * we ask Yahoo about the best zoom level to display this area on 1000x800 and
#   request the image
# * we ask Yahoo about the extents actually displayed and receive the answer:
#   (48.45,6.95)-(48.85,7.15)
# * This is .4 degrees high (we requested .3 degrees high) and .2 degrees wide
#   (we requested .1 wide).
# * We scale the image to 1333x1600 and cut a 1000x800 section from the middle,
#   knowing that this will now be exactly .3 degrees by .1 degrees (minus projection
#   errors of course!)
#
#
# Author: Frederik Ramm <frederik@remote.org>
# License: Public Domain
#
# Version History
# ---------------
# 0.1 (12 Jan 2007) initial publication
#
package YWMS;

use base qw(HTTP::Server::Simple::CGI);
use strict;

# called by HTTP::Server::Simple to process requests. 
# (does not return befor request is completed, i.e. only one request
# can be processed at any one time)

sub handle_request 
{

    my ($self, $cgi) = @_;

    # very simple, we don't even look at the name of the requested CGI, we just
    # look at some parameters

    if ($cgi->param("request") ne "GetMap")
    {
        print "500 HTTP/1.1 Internal Server Error\n";
        print "Content-type: text/html\n\n<h1>Internal Server Error</h1>";
        return;
    }

    # fork; parent waits, child works. this is done to return to a 100% clean
    # environment after the request is processed.
   
    my $forkresult  = fork();
    if ($forkresult < 0)
    {
        print "500 HTTP/1.1 Internal Server Error\n";
        print "Content-type: text/html\n\n<h1>Internal Server Error</h1>";
        return;
    }
    elsif ($forkresult > 0)
    {
        waitpid($forkresult, 0);
        return;
    }

    my ($bbox, $width, $height) = ($cgi->param("bbox"), $cgi->param("width"), $cgi->param("height"));
    my ($tllon, $tllat, $brlon, $brlat) = split(/,/, $bbox);

    my ($atllat,$atllon, $abrlat, $abrlon); 

    printf STDERR "top-left lat %f\n", $tllat;
    printf STDERR "top-left lon %f\n", $tllon;
    printf STDERR "bot-right lat %f\n", $brlat;
    printf STDERR "bot-right lon %f\n", $brlon;
    printf STDERR "width %d\n", $width;
    printf STDERR "height %d\n", $height;

    # change this to match your firefox setup. 
    #my ($xcut, $ycut) = (9, 111); iceape
    my ($xcut, $ycut) = (8, 101);
    my $browser = "firefox";

    my $tmpdir = "/tmp/ywms.$$";
    mkdir $tmpdir or die;

    # create "web page" that will show the image we want
    open (TMP, "> $tmpdir/src.html") or die;

    print TMP <<EOF;
<html>
<head>
<script type="text/javascript">
function fullscreen()
{
    window.moveTo(0,0);
    window.resizeTo(screen.width,screen.height);
    // this "finished" is required for the waiting web server process to recognize that
    // the web page is fully rendered. If it doesn't work for some reason, can also be
    // replaced by a "sleep" command in the server.
    dump(">>finished\\n");
}
</script>
<script type="text/javascript" src="http://api.maps.yahoo.com/ajaxymap?v=3.0&appid=YahooDemo"></script>

</head>
<body onload="fullscreen();">
  
<div id="map" style="width: ${width}px; height: ${height}px"></div>
<p>
<div id="llb"></div>
<script type="text/javascript">
// other map types work as well but only satellite allowed for use with OSM!
var map = new YMap(document.getElementById("map"), YAHOO_MAP_SAT);
var points = new Array();
points[0] = new YGeoPoint($tllat, $tllon);
points[1] = new YGeoPoint($brlat, $brlon);

var zac = map.getBestZoomAndCenter(points);
var level = zac.zoomLevel;
// funny Yahoo bug seems to return 0 if your section is too small
if (level==0) level=1;
map.drawZoomAndCenter(zac.YGeoPoint,level);
var bounds = map.getBoundsLatLon();
// the format of this has to match the RegEx in the server, see further down
dump(">>actual:((");
dump(bounds.LatMin+", "+bounds.LonMin+"), ("+bounds.LatMax+", "+bounds.LonMax+"))\\n");
</script>

</body>
</html>
EOF

    close(TMP);

    # change this if your X display has another number
    my $xdisplay = 1;
    my $xscreen = 0;

    # fork the virtual frame buffer
    my $xvfbpid = fork();
    if ($xvfbpid == 0)
    {
        # allow 300 extra pixels for browser frills
        # if desired change the display colour depth here (x8, x16, x24)
        exec("Xvfb", ":$xdisplay", "-screen",$xscreen,sprintf("%dx%dx8", $width+300,$height+300), "-fbdir", $tmpdir) or die;
    }
    elsif ($xvfbpid < 0) 
    {
        die;
    }

    # sleep(1);
    
    # try to move mouse cursor to 0,0 - does not seem to work?
    system("xwit -display :$xdisplay.$xscreen -root -warp 0 0");

    # start the browser. width and height probably unnecessary as long as fullscreen resize
    # in javascript code is working!
    my $browserpid = open(BROWSER,sprintf("$browser -no-remote --display=:$xdisplay.$xscreen -width %d -height %d file://$tmpdir/src.html|", $width+300,$height+300)) or die;

    # wait for "finished" message from browser.
    while(<BROWSER>)
    {
        print STDERR "++ $_";
        if (/^>>actual:\(\((.*),(.*)\), \((.*),(.*)\)\)/)
        {
            ($atllat, $atllon, $abrlat, $abrlon) = ($1, $2, $3, $4);
        }
        elsif (/^>>finished/)
        {
            last;
        }
    }

    # do NOT "close BROWSER" here as this implies a waitpid. we have to kill
    # the browser without closing. don't bother about the wasted file descriptor,
    # the process will die anyway.

    printf STDERR "actual top-left lat %f\n", $atllat;
    printf STDERR "actual top-left lon %f\n", $atllon;
    printf STDERR "actual bot-right lat %f\n", $abrlat;
    printf STDERR "actual bot-right lon %f\n", $abrlon;

    # calculate resize and cropping (as described in the opening comment)
    
    my $calcheight = ($abrlat - $atllat) * $height / ($brlat - $tllat);
    my $calcwidth = ($abrlon - $atllon) * $width / ($brlon - $tllon);

    my $cropl = ($calcwidth - $width) / 2;
    my $cropt = ($calcheight - $height) / 2;

    printf STDERR "calculated width/height: %d / %d\n", $calcwidth, $calcheight;

    # you may add extra stuff into the following pipe to change the image, e.g. make it lighter
    
    my $command = "xwdtopnm < $tmpdir/Xvfb_screen$xscreen | ".
                  "pnmcut -left $xcut -top $ycut -width $width -height $height | ".
                  sprintf("pnmscale -width %d -height %d |", $calcwidth, $calcheight).
                  sprintf("pnmcut -left %d -top %d -width %d -height %d |", $cropl, $cropt, $width, $height).
                  "cjpeg > $tmpdir/image.jpg\n";

    print STDERR $command;
    system("/bin/sh", "-c", $command);

    # dispose of child processes 

    kill 9, $browserpid;
    kill 9, $xvfbpid;

    # return data
    
    print("HTTP/1.1 200 OK\n");
    print("Content-type: image/jpeg\n\n");
    undef $/;
    open(IMG, "$tmpdir/image.jpg");
    my $i = <IMG>;
    close(IMG);
    print $i;

    # remove temporary files

    unlink "$tmpdir/image.jpg";
    unlink "$tmpdir/Xvfb_screen0";
    unlink "$tmpdir/src.html";
    rmdir $tmpdir;
    exit;

}
 
1;
